from bisect import insort
from collections import OrderedDict


class QCMetric(object):
    """Container that holds the qc metric as OrderedDict (sorted by keys of
    the input dict) and the "master" key (name) of the said qc metric. Can be
    instantiated from a regular dict.
    """

    def __init__(self, qc_metric_name, qc_metric_content):
        qc_metric_dict = qc_metric_content
        if not isinstance(qc_metric_dict, dict):
            raise TypeError("QCMetric data must be a dict.")
        self._name = qc_metric_name
        self._content = OrderedDict(sorted(qc_metric_dict.items(), key=lambda x: x[0]))

    @property
    def content(self):
        return self._content

    @property
    def name(self):
        return self._name

    def to_ordered_dict(self):
        """Returns an OrderedDict with the contents.
        Returns: OrderedDict with structure:
            {self._name: self._content}
        """
        return OrderedDict([(self._name, self._content)])

    def __len__(self):
        return len(self._content)

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return "QCMetric(%r, %r)" % (self.name, self.content)


class QCMetricRecord(object):
    """Container that holds QCMetrics in sorted order.

    Attributes:
        metrics: list of metrics, kept sorted by the name of metrics
    """

    def __init__(self, metrics=None, name=None):
        if metrics is None:
            self._metrics = []
        else:
            # names must be unique
            names = [metric.name for metric in metrics]
            assert len(names) == len(set(names)), "Names of metrics have to be unique"
            self._metrics = metrics[:]
            self._metrics.sort()
        self._name = name

    @property
    def metrics(self):
        return self._metrics

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def add(self, qc_metric):
        """Adds qc metric to the metrics, keeping it sorted by name.

        Args:
            qc_metric: QCMetric

        Returns: None

        Raises: AssertionError if a metric with same name is already in record
        """

        assert (
            qc_metric not in self._metrics
        ), "Metric with name {} already in record".format(qc_metric.name)
        insort(self._metrics, qc_metric)

    def add_all(self, qc_metric_container):
        """Adds all metrics from qc_metric_container preserving uniquness of names and order.
        If the names in the qc_metric_container are not unique, raises AssertionError and leaves
        the QCMetricRecord unmodified.
        Args:
            qc_metric_container: container of QCMetrics
        Returns: None

        Raises: AssertionError if adding would break the uniqueness of names in self.
        """
        for metric in qc_metric_container:
            assert (
                metric not in self._metrics
            ), "Metric with name {} already in record. Nothing from the container added".format(
                metric.name
            )
        for metric in qc_metric_container:
            self.add(metric)

    def to_ordered_dict(self):
        """Returns an OrderedDict with the contents.

        Returns: Ordered dict with structure as follows:
            - Ordered as the metrics is
            - Contents, assuming metrics = [qc1, qc2, qc3]:
            {
                qc1.name : qc1.content,
                qc2.name : qc2.content,
                qc3.name : qc3.content
            }
            If the self._name is set, then the above output will be changed into
            {
                self._name : {
                                qc1.name : qc1.content,
                                qc2.name : qc2.content,
                                qc3.name : qc3.content
                             }
            }
        """
        result = OrderedDict()
        for metric in self._metrics:
            result.update({metric.name: metric.content})
        if self._name is not None:
            result = OrderedDict([(self._name, result)])
        return result

    def __len__(self):
        """
        Delegated to metrics.
        """
        return len(self._metrics)

    def __iter__(self):
        """
        Iterating QCMetricRecord is iterating over metrics.
        """
        return iter(self._metrics)

    def __repr__(self):
        """
        Like __iter__, __repr__ is delegated to metrics.
        """
        if self._name is not None:
            name_repr_token = ", name='%s'" % self._name
        else:
            name_repr_token = ""
        return "QCMetricRecord(%s%s)" % (self._metrics.__repr__(), name_repr_token)



# Parser functions


def parse_starlog(path_to_log):
    """Parse star logfile into a dict.
    Args:
        path_to_log: filepath containing the star run log
    Returns:
        dict that contains metrics from star log
    """
    qc_dict = {}
    with open(path_to_log) as f:
        for line in f:
            if "|" in line:
                tokens = line.split("|")
                qc_dict[tokens[0].strip()] = tokens[1].strip()
    return qc_dict


# format pergentage points


def percentage_to_float(line):
    return float(line.strip("%"))


def try_converting_to_numeric(value):
    """Try converting to numeric.
    Attempts to convert to int, and if that fails, to float
    and as a last resort keep as str.
    Args:
        value: str input value from qc
    Returns:
        value converted to numeric if applicable.
    """
    assert type(value) == str
    try:
        converted_value = int(value)
    except ValueError:
        try:
            converted_value = float(value)
        except ValueError:
            converted_value = value
    return converted_value


def parse_flagstats(filePath):
    """Parse samtools flagstat file into a dict
    Args:
        filePath: path to samtools flagstat file
    Returns:
        dict that contains metrics from flagstats
    """
    pairs = {}
    numbers_type = int
    with open(filePath, "r") as fh:
        while True:
            line = fh.readline()
            if line is None:
                break
            line.strip()
            if line == "":
                continue
            # 2826233 + 0 in total (QC-passed reads + QC-failed reads)
            if line.find("QC-passed reads") > 0:
                # 2826233 + 0 in total (QC-passed reads + QC-failed reads)
                parts = line.split()
                pairs["total"] = numbers_type(parts[0])
                pairs["total_qc_failed"] = numbers_type(parts[2])
            # 0 + 0 duplicates
            elif line.find("duplicates") > 0:
                parts = line.split()
                pairs["duplicates"] = numbers_type(parts[0])
                pairs["duplicates_qc_failed"] = numbers_type(parts[2])
            # 2826233 + 0 mapped (100.00%:-nan%)
            elif "mapped" not in pairs and line.find("mapped") > 0:
                parts = line.split()
                pairs["mapped"] = numbers_type(parts[0])
                pairs["mapped_qc_failed"] = numbers_type(parts[2])
                val = parts[4][1:].split(":")[0]
                pairs["mapped_pct"] = percentage_to_float(val)
            # 2142 + 0 paired in sequencing
            elif line.find("paired in sequencing") > 0:
                parts = line.split()
                if int(parts[0]) <= 0:  # Not paired-end, so nothing more needed
                    break
                pairs["paired"] = numbers_type(parts[0])
                pairs["paired_qc_failed"] = numbers_type(parts[2])
            # 107149 + 0 read1
            elif line.find("read1") > 0:
                parts = line.split()
                pairs["read1"] = numbers_type(parts[0])
                pairs["read1_qc_failed"] = numbers_type(parts[2])
            # 107149 + 0 read2
            elif line.find("read2") > 0:
                parts = line.split()
                pairs["read2"] = numbers_type(parts[0])
                pairs["read2_qc_failed"] = numbers_type(parts[2])
            # 2046 + 0 properly paired (95.48%:-nan%)
            elif line.find("properly paired") > 0:
                parts = line.split()
                pairs["paired_properly"] = numbers_type(parts[0])
                pairs["paired_properly_qc_failed"] = numbers_type(parts[2])
                val = parts[5][1:].split(":")[0]
                pairs["paired_properly_pct"] = percentage_to_float(val)
            # 0 + 0      singletons (0.00%:-nan%)
            elif line.find("singletons") > 0:
                parts = line.split()
                pairs["singletons"] = numbers_type(parts[0])
                pairs["singletons_qc_failed"] = numbers_type(parts[2])
                val = parts[4][1:].split(":")[0]
                pairs["singletons_pct"] = percentage_to_float(val)
            # 2046212 + 0 with itself and mate mapped
            elif line.find("with itself and mate mapped") > 0:
                parts = line.split()
                pairs["with_itself"] = numbers_type(parts[0])
                pairs["with_itself_qc_failed"] = numbers_type(parts[2])
            # 0 + 0 with mate mapped to a different chr (mapQ>=5)
            elif line.find("with mate mapped to a different chr") > 0:
                parts = line.split()
                pairs["diff_chroms"] = numbers_type(parts[0])
                pairs["diff_chroms_qc_failed"] = numbers_type(parts[2])
                break
    return pairs


def parse_key_value_tsv(path_to_tsv, sep=None):
    """Parse 2-column tsv file with string key, valuetype value into a dict.
    Args:
        path_to_tsv: path to file containing the tsv
        sep: column separator String, default: None (will split on any whitespace)
    Raises:
        AssertionError if a row that splits into number of tokens other than 2
    """
    qc_dict = {}
    with open(path_to_tsv) as f:
        for line in f:
            tokens = line.split(sep)
            assert len(tokens) == 2
            qc_dict[tokens[0]] = try_converting_to_numeric(tokens[1])
    return qc_dict


def parse_illumina_trimstats(path_to_trimstats):
    """Parse illumina trimstats into a dict.
    Args:
        path_to_trimstats: path to file that contains illumina trimstats
    Returns:
        dict with trimstats qc metrics
    Raises:
        AssertionError if a row that splits into number of tokens other than 2
        ValueError if some value is not suitable for converting into int
    """
    qc_dict = {}
    with open(path_to_trimstats) as f:
        rawstats = f.readline()
    tokens = rawstats.split(";")
    qc_dict["Total read-pairs processed"] = int(tokens[0].split()[-1])
    qc_dict["Total read-pairs trimmed"] = int(tokens[1].split()[-1])
    return qc_dict


def parse_bamcounts(path_to_bamcounts):
    """Parse bamcounts into a dict.
    Args:
        path_to_bamcounts: path to file that contains bamcounts/tagcounts
    Returns:
        dict with bamcounts qc metrics
    Raises:
        AssertionError if a row that splits into number of tokens other than 2
    """
    return parse_key_value_tsv(path_to_bamcounts)


def parse_preseq_targets(path_to_preseq_targets):
    """Parse preseq targets into a dict.
    Args:
        path_to_preseq_targets: path to file that contains preeq targets
    Returns:
        dict with preseq targets qc metrics
    Raises:
        AssertionError if a row that splits into number of tokens other than 2
    """
    return parse_key_value_tsv(path_to_preseq_targets)


def parse_hotspot1_spot_score(path_to_hotspot1_spot_score):
    """Parse hotspot1 spot score into a dict.
    Args:
        path_to_hotspot1_spot_score: path to file that contains hotspot1 spot score
    Returns:
        dict with hotspot1 spot score qc metrics
    """
    qc_dict = {}
    with open(path_to_hotspot1_spot_score) as f:
        next(f)
        raw_values = f.readline()
    values = raw_values.strip().split()
    qc_dict["total tags"] = try_converting_to_numeric(values[0])
    qc_dict["hotspot tags"] = try_converting_to_numeric(values[1])
    qc_dict["SPOT"] = try_converting_to_numeric(values[2])
    return qc_dict


def parse_picard_metrics(path_to_metrics):
    """Parse picard metrics into a dict.
    Args:
        path_to_metrics: path to a file that contains picard etrics
    Returns:
        dict with picard metrics
    """
    qc_dict = {}
    with open(path_to_metrics) as f:
        lines = []
        for _ in range(8):
            lines.append(f.readline())
    keys = lines[6].split("\t")
    values = lines[7].split("\t")
    keys = [key.strip() for key in keys]
    values = [value.strip() for value in values]
    for key, value in zip(keys, values):
        qc_dict[key] = try_converting_to_numeric(value)
    return qc_dict


def parse_picard_duplication_metrics(path_to_duplication_metrics):
    """Parse picard duplication metrics into a dict.
    Args:
        path_to_duplication_metrics: path to a file that contains picard duplication metrics
    Returns:
        dict with picard duplication metrics
    """
    return parse_picard_metrics(path_to_duplication_metrics)


def parse_picard_insert_size_metrics(path_to_insert_size_metrics):
    """Parse picard insert size metrics into a dict.
    Args:
        path_to_insert_size_metrics: path to a file that contains picard insert size metrics
    Returns:
        dict with picard insert size metrics
    """
    return parse_picard_metrics(path_to_insert_size_metrics)


def parse_insert_size_info(path_to_insert_size_info):
    return parse_key_value_tsv(path_to_insert_size_info)


def parse_samtools_stats(path_to_samtools_stats):
    """Parse samtools stats metrics into a dict.
    Args:
        path_to_samtools_stats: path to a file that contains samtools stats metrics
    Returns:
        dict with samtools stats metrics
    """
    with open(path_to_samtools_stats) as f:
        stats_lines = [
            line.split("\t") for line in f.readlines() if line.startswith("SN")
        ]
    qc_dict = {
        line[1].strip(":"): try_converting_to_numeric(line[2]) for line in stats_lines
    }
    return qc_dict



from collections import OrderedDict

import pytest

from qc_utils import QCMetric, QCMetricRecord


@pytest.fixture
def ordered():
    return OrderedDict([(2, "a"), (1, "b")])


@pytest.fixture
def obj_a1():
    return QCMetric("a", {1: 2})


@pytest.fixture
def obj_a2():
    return QCMetric("a", {2: 3})


@pytest.fixture
def obj_b():
    return QCMetric("b", {3: 4})


@pytest.fixture
def obj_c():
    return QCMetric("c", {1: 2, 3: 4, 5: 6})


@pytest.fixture
def obj_d():
    return QCMetric("d", {"a": "b"})


@pytest.fixture
def qc_record():
    return QCMetricRecord()


# Test QCMetric


def test_type_check():
    with pytest.raises(TypeError):
        QCMetric("name", 1)


def test_get_name():
    qc_obj = QCMetric("a", {})
    assert qc_obj.name == "a"


def test_get_content():
    qc_obj = QCMetric("_", {2: "a", 1: "b"})
    assert qc_obj.content == OrderedDict([(1, "b"), (2, "a")])


def test_len_0():
    assert len(QCMetric("a", {})) == 0


def test_len_1(obj_a1):
    assert len(obj_a1) == 1


def test_metric_to_ordered_dict(obj_d):
    assert obj_d.to_ordered_dict() == OrderedDict([("d", OrderedDict([("a", "b")]))])


def test_less_than():
    smaller_obj = QCMetric(1, {})
    bigger_obj = QCMetric(2, {})
    assert smaller_obj < bigger_obj


def test_equals():
    first_obj = QCMetric("a", {})
    second_obj = QCMetric("a", {"x": "y"})
    assert first_obj == second_obj


def test_QCMetric_repr():
    obj = QCMetric("a", {1: "x"})
    assert obj.__repr__() == "QCMetric('a', OrderedDict([(1, 'x')]))"


# Test QCMetricRecord


def test_init_from_list_not_unique(obj_a1, obj_a2):
    metrics = [obj_a1, obj_a2]
    with pytest.raises(AssertionError):
        QCMetricRecord(metrics)


def test_init_from_list_success(obj_a1, obj_b):
    metrics = [obj_a1, obj_b]
    record = QCMetricRecord(metrics)
    assert record.metrics[0] is obj_a1
    assert record.metrics[1] is obj_b


def test_add(qc_record, obj_a1):
    assert len(qc_record) == 0
    qc_record.add(obj_a1)
    assert len(qc_record) == 1


def test_add_all_to_empty(qc_record, obj_a1, obj_b):
    metrics = [obj_a1, obj_b]
    qc_record.add_all(metrics)
    assert len(qc_record) == 2


def test_add_all_to_nonempty_success(qc_record, obj_a1, obj_b, obj_c, obj_d):
    metrics = [obj_a1, obj_b]
    record = QCMetricRecord(metrics)
    record.add_all([obj_c, obj_d])
    assert len(record) == 4


def test_add_all_failure_because_not_unique(obj_a1, obj_a2, obj_b):
    record = QCMetricRecord([obj_a1])
    with pytest.raises(AssertionError):
        record.add_all([obj_b, obj_a2])
    assert len(record) == 1


def test_add_raises_error_when_add_same_twice(qc_record, obj_a1):
    qc_record.add(obj_a1)
    with pytest.raises(AssertionError):
        qc_record.add(obj_a1)


def test_add_raises_error_when_add_with_same_name(qc_record, obj_a1, obj_a2):
    qc_record.add(obj_a1)
    with pytest.raises(AssertionError):
        qc_record.add(obj_a2)


def test_to_ordered_dict(qc_record, obj_a1, obj_b):
    qc_record.add(obj_a1)
    qc_record.add(obj_b)
    qc_dict = qc_record.to_ordered_dict()
    assert qc_dict == OrderedDict([("a", {1: 2}), ("b", {3: 4})])


def test_QCMetricRecord_repr(obj_a1, obj_b):
    metrics = [obj_a1, obj_b]
    record = QCMetricRecord(metrics)
    assert (
        record.__repr__()
        == "QCMetricRecord([QCMetric('a', OrderedDict([(1, 2)])), QCMetric('b', OrderedDict([(3, 4)]))])"
    )


def test_named_QCMetricRecord_repr(obj_a1, obj_b):
    metrics = [obj_a1, obj_b]
    record = QCMetricRecord(metrics, name="dis_my_name")
    assert (
        record.__repr__()
        == "QCMetricRecord([QCMetric('a', OrderedDict([(1, 2)])), QCMetric('b', OrderedDict([(3, 4)]))], name='dis_my_name')"
    )


def test_QCMetricRecord_getname():
    named_record = QCMetricRecord(name="dis_my_name")
    assert named_record.name == "dis_my_name"


def test_QCMetricRecord_setname(qc_record):
    assert qc_record.name is None
    qc_record.name = "dis_my_name"
    assert qc_record.name == "dis_my_name"


def test_to_ordered_dict_name_is_set(qc_record, obj_a1, obj_b):
    qc_record.add(obj_a1)
    qc_record.add(obj_b)
    qc_record.name = "dis_my_name"
    qc_dict = qc_record.to_ordered_dict()
    assert qc_dict == OrderedDict(
        [("dis_my_name", OrderedDict([("a", {1: 2}), ("b", {3: 4})]))]
    )



from qc_utils.qcmetric import QCMetric, QCMetricRecord

__title__ = "qc-utils"
__version__ = "20.05.1"
__description__ = "Tool for representing ENCODE QC metrics"
__url__ = "https://github.com/ENCODE-DCC/qc-utils"
__uri__ = __url__
__author__ = "Otto Jolanki"
__email__ = "ojolanki@stanford.edu"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2020 ENCODE DCC"



from io import StringIO
from unittest.mock import patch

import pytest

from qc_utils import parsers, qcmetric

STAR_LOG = """                                 Started job on | Feb 16 23:45:04
                         Started mapping on |   Feb 16 23:49:02
                                Finished on |   Feb 17 00:16:34
   Mapping speed, Million of reads per hour |   115.09

                      Number of input reads |   52815760
                  Average input read length |   202
                                UNIQUE READS:
               Uniquely mapped reads number |   49542908
                    Uniquely mapped reads % |   93.80%
                      Average mapped length |   201.00
                   Number of splices: Total |   20633562
        Number of splices: Annotated (sjdb) |   20453258
                   Number of splices: GT/AG |   20442117
                   Number of splices: GC/AG |   154577
                   Number of splices: AT/AC |   11212
           Number of splices: Non-canonical |   25656
                  Mismatch rate per base, % |   0.30%
                     Deletion rate per base |   0.02%
                    Deletion average length |   1.51
                    Insertion rate per base |   0.01%
                   Insertion average length |   1.37
                         MULTI-MAPPING READS:
    Number of reads mapped to multiple loci |   2218531
         % of reads mapped to multiple loci |   4.20%
    Number of reads mapped to too many loci |   7303
         % of reads mapped to too many loci |   0.01%
                              UNMAPPED READS:
   % of reads unmapped: too many mismatches |   0.00%
             % of reads unmapped: too short |   1.95%
                 % of reads unmapped: other |   0.03%
                              CHIMERIC READS:
                   Number of chimeric reads |   0
                        % of chimeric reads |   0.00%"""

SAMTOOLS_FLAGSTAT = """424886248 + 0 in total (QC-passed reads + QC-failed reads)
0 + 0 duplicates
413471158 + 0 mapped (97.31%:-nan%)
424886248 + 0 paired in sequencing
212443124 + 0 read1
212443124 + 0 read2
413471158 + 0 properly paired (97.31%:-nan%)
413471158 + 0 with itself and mate mapped
0 + 0 singletons (0.00%:-nan%)
0 + 0 with mate mapped to a different chr
0 + 0 with mate mapped to a different chr (mapQ>=5)"""

ILLUMINA_TRIMSTATS = (
    "Total read-pairs processed: 179710367; Total read-pairs trimmed: 8621991"
)

BAMCOUNTS = """paired-nuclear-align 208933956
qc-flagged 0
u 263952413
u-pf 263952413
u-pf-n 259418766
u-pf-n-mm2 259189714
u-pf-n-mm2-mito  240666871
chr1 22465899
chr10 11223351
chr11 12368547
chr11_KI270721v1_random 422
chr12 11792546
chr13 6786761
chr14 8023861
chr14_GL000009v2_random 2699
"""

PRESEQ_TARGETS = """preseq-est-for-20000000 21000000
preseq-est-for-30000000  31000000
preseq-est-for-40000000  42000000
preseq-est-for-100000000 107000000
preseq-est-for-150000000 165000000
preseq-est-for-250000000 285000000
preseq-est-max 752199191
"""
HOTSPOT1_SPOT_SCORE = """  total tags  hotspot tags    SPOT
     5000000       2209334  0.4418\n"""

PICARD_DUPLICATION_METRICS = """## htsjdk.samtools.metrics.StringHeader
# picard.sam.markduplicates.MarkDuplicatesWithMateCigar MINIMUM_DISTANCE=300 INPUT=[/cromwell_root/dnase/dnase/a707c686-ca59-4eed-91f1-a287f72154a0/call-dnase_replicate/shard-0/dnase_replicate/6f9d8f96-9920-4930-940b-9876fcded62a/call-merge_mark_and_filter_bams/merge_mark_and_filter_bams/25a1ee0c-184e-4526-bf62-c55c16633184/call-mark/mark/85f77336-810b-40fb-abd3-f0c72201fc4e/call-add_mate_cigar_to_bam/add_mate_cigar_to_bam/7f0df874-11a9-47b1-93fd-c4a98f967638/call-revert_original_base_qualities_and_add_mate_cigar/mate_cigar.bam] OUTPUT=marked.bam METRICS_FILE=MarkDuplicates.picard ASSUME_SORTED=true READ_NAME_REGEX=[a-zA-Z0-9]+:[0-9]+:[a-zA-Z0-9]+:[0-9]+:([0-9]+):([0-9]+):([0-9]+).* VALIDATION_STRINGENCY=SILENT    SKIP_PAIRS_WITH_NO_MATE_CIGAR=true BLOCK_SIZE=100000 REMOVE_DUPLICATES=false DUPLICATE_SCORING_STRATEGY=TOTAL_MAPPED_REFERENCE_LENGTH PROGRAM_RECORD_ID=MarkDuplicates PROGRAM_GROUP_NAME=MarkDuplicatesWithMateCigar OPTICAL_DUPLICATE_PIXEL_DISTANCE=100 VERBOSITY=INFO QUIET=false COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false GA4GH_CLIENT_SECRETS=client_secrets.json
## htsjdk.samtools.metrics.StringHeader
# Started on: Mon May 04 00:58:33 UTC 2020

## METRICS CLASS picard.sam.DuplicationMetrics
LIBRARY	UNPAIRED_READS_EXAMINED	READ_PAIRS_EXAMINED	SECONDARY_OR_SUPPLEMENTARY_RDS	UNMAPPED_READS	UNPAIRED_READ_DUPLICATES	READ_PAIR_DUPLICATES	READ_PAIR_OPTICAL_DUPLICATES	PERCENT_DUPLICATION	ESTIMATED_LIBRARY_SIZE
Unknown Library	54238976	137746811	0	16929897	22741788	17466587	0	0.174914	496198880

## HISTOGRAM	java.lang.Double
"""

PICARD_INSERT_SIZE_METRICS = """## htsjdk.samtools.metrics.StringHeader
# picard.analysis.CollectInsertSizeMetrics HISTOGRAM_FILE=nuclear.CollectInsertSizeMetrics.picard.pdf INPUT=/cromwell_root/dnase/dnase/1b6c015b-4fa9-444d-864d-285c188c0b8f/call-dnase_replicate/shard-0/dnase_replicate/ab6fb07f-7419-4688-8e84-87e869a51ea5/call-merge_mark_and_filter_bams/merge_mark_and_filter_bams/1b2ff752-be3c-4544-8a33-950adf5a59ef/call-filter/filter/f39085ca-823f-4d64-bfbb-34d672980f20/call-filter_bam_reads_with_nonnuclear_flag/filter_bam_reads_with_nonnuclear_flag/e99d92cb-cb88-468b-a360-659c39a6a01b/call-view/nuclear.bam OUTPUT=nuclear.CollectInsertSizeMetrics.picard ASSUME_SORTED=true VALIDATION_STRINGENCY=LENIENT    DEVIATIONS=10.0 MINIMUM_PCT=0.05 METRIC_ACCUMULATION_LEVEL=[ALL_READS] INCLUDE_DUPLICATES=false STOP_AFTER=0 VERBOSITY=INFO QUIET=false COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false GA4GH_CLIENT_SECRETS=client_secrets.json
## htsjdk.samtools.metrics.StringHeader
# Started on: Thu Apr 23 17:17:19 UTC 2020

## METRICS CLASS	picard.analysis.InsertSizeMetrics
MEDIAN_INSERT_SIZE	MEDIAN_ABSOLUTE_DEVIATION	MIN_INSERT_SIZE	MAX_INSERT_SIZE	MEAN_INSERT_SIZE	STANDARD_DEVIATION	READ_PAIRS	PAIR_ORIENTATION	WIDTH_OF_10_PERCENT	WIDTH_OF_20_PERCENT	WIDTH_OF_30_PERCENT	WIDTH_OF_40_PERCENT	WIDTH_OF_50_PERCENT	WIDTH_OF_60_PERCENT	WIDTH_OF_70_PERCENT	WIDTH_OF_80_PERCENT	WIDTH_OF_90_PERCENT	WIDTH_OF_99_PERCENT	SAMPLE	LIBRARY	READ_GROUP
82	6	5	750	84.615764	11.674485	13631725	FR	3	7	9	11	13	17	19	31	93	517

## HISTOGRAM	java.lang.Integer
"""

SAMTOOLS_STATS = """CHK	b4f24be5	5ef36d97	450ccdce
# Summary Numbers. Use `grep ^SN | cut -f 2-` to extract this part.
SN	raw total sequences:	137377422
SN	filtered sequences:	0
SN	sequences:	137377422
SN	is sorted:	0
SN	1st fragments:	137377422
SN	last fragments:	0
SN	reads mapped:	133755927
SN	reads mapped and paired:	0	# paired-end technology bit set + both mates mapped
SN	reads unmapped:	3621495
SN	reads properly paired:	0	# proper-pair bit set
SN	reads paired:	0	# paired-end technology bit set
SN	reads duplicated:	0	# PCR or optical duplicate bit set
SN	reads MQ0:	27952198	# mapped and MQ=0
SN	reads QC failed:	0
SN	non-primary alignments:	0
SN	total length:	4945587192	# ignores clipping
SN	total first fragment length:	4945587192	# ignores clipping
SN	total last fragment length:	0	# ignores clipping
SN	bases mapped:	4815213372	# ignores clipping
SN	bases mapped (cigar):	4815213372	# more accurate
SN	bases trimmed:	0
SN	bases duplicated:	0
SN	mismatches:	15823799	# from NM fields
SN	error rate:	3.286209e-03	# mismatches / bases mapped (cigar)
SN	average length:	36
SN	average first fragment length:	36
SN	average last fragment length:	0
SN	maximum length:	36
SN	maximum first fragment length:	36
SN	maximum last fragment length:	0
SN	average quality:	65.8
SN	insert size average:	0.0
SN	insert size standard deviation:	0.0
SN	inward oriented pairs:	0
SN	outward oriented pairs:	0
SN	pairs with other orientation:	0
SN	pairs on different chromosomes:	0
SN	percentage of properly paired reads (%):	0.0
# First Fragment Qualities. Use `grep ^FFQ | cut -f 2-` to extract this part.
"""

INSERT_SIZE_INFO = """insert-ls-ratio	2983.1667
insert-ft-eleven	0.0912
"""


@patch("builtins.open", return_value=StringIO(STAR_LOG))
def test_parse_starlog(mock_open):
    star_log_dict = parsers.parse_starlog("path")
    assert len(star_log_dict) == 29


def test_percentage_to_float():
    percentage_line = "94.67%"
    formatted = parsers.percentage_to_float(percentage_line)
    assert isinstance(formatted, float)
    assert formatted == 94.67


@patch("builtins.open", return_value=StringIO(SAMTOOLS_FLAGSTAT))
def test_parse_flagstats(mock_open):
    flagstat_dict = parsers.parse_flagstats("path")
    assert len(flagstat_dict) == 23


@patch("builtins.open", return_value=StringIO(ILLUMINA_TRIMSTATS))
def test_parse_illumina_trimstats(mock_open):
    trimstat_dict = parsers.parse_illumina_trimstats("path")
    assert len(trimstat_dict) == 2
    assert trimstat_dict["Total read-pairs processed"] == 179710367
    assert trimstat_dict["Total read-pairs trimmed"] == 8621991


@patch("builtins.open", return_value=StringIO(BAMCOUNTS))
def test_parse_bamcounts(mock_open):
    bamstat_dict = parsers.parse_bamcounts("path")
    assert len(bamstat_dict) == 15
    assert "u-pf-n-mm2-mito" in bamstat_dict
    assert bamstat_dict["chr14_GL000009v2_random"] == 2699


@patch("builtins.open", return_value=StringIO("chr1 123 456"))
def test_parse_bamcounts_bad_input_raises_AssertionError(mock_open):
    with pytest.raises(AssertionError):
        parsers.parse_bamcounts("path")


@patch("builtins.open", return_value=StringIO(PRESEQ_TARGETS))
def test_parse_preseq_targets(mock_open):
    preseq_targets_dict = parsers.parse_preseq_targets("path")
    assert len(preseq_targets_dict) == 7
    assert "preseq-est-for-100000000" in preseq_targets_dict
    assert preseq_targets_dict["preseq-est-max"] == 752199191


@patch("builtins.open", return_value=StringIO(HOTSPOT1_SPOT_SCORE))
def test_parse_hotspot1_spot_score(mock_open):
    spot_score_dict = parsers.parse_hotspot1_spot_score("path")
    assert len(spot_score_dict) == 3
    assert spot_score_dict["hotspot tags"] == 2209334
    assert spot_score_dict["SPOT"] == 0.4418


@patch("builtins.open", return_value=StringIO(PICARD_DUPLICATION_METRICS))
def test_parse_picard_duplication_metrics(mock_open):
    duplication_metrics_dict = parsers.parse_picard_duplication_metrics("path")
    assert len(duplication_metrics_dict) == 10
    assert "UNMAPPED_READS" in duplication_metrics_dict
    assert duplication_metrics_dict["PERCENT_DUPLICATION"] == 0.174914


def test_try_converting_to_numeric_raises():
    with pytest.raises(AssertionError):
        parsers.try_converting_to_numeric(3.14)


def test_try_converting_to_numeric_int():
    assert parsers.try_converting_to_numeric("1") == 1


def test_try_converting_to_numeric_float():
    assert parsers.try_converting_to_numeric("3.14") == 3.14


def test_try_converting_to_numeric_str():
    assert parsers.try_converting_to_numeric("SPAM") == "SPAM"


@patch("builtins.open", return_value=StringIO(INSERT_SIZE_INFO))
def test_parse_insert_size_info(mock_open):
    insert_size_info_dict = parsers.parse_insert_size_info("path")
    assert len(insert_size_info_dict) == 2
    assert "insert-ft-eleven" in insert_size_info_dict
    assert insert_size_info_dict["insert-ls-ratio"] == 2983.1667


@patch("builtins.open", return_value=StringIO(PICARD_INSERT_SIZE_METRICS))
def test_parse_picard_insert_size_metrics(mock_open):
    insert_size_metrics_dict = parsers.parse_picard_insert_size_metrics("path")
    assert len(insert_size_metrics_dict) == 18
    assert "READ_PAIRS" in insert_size_metrics_dict
    assert insert_size_metrics_dict["MEAN_INSERT_SIZE"] == 84.615764
    assert insert_size_metrics_dict["MEDIAN_INSERT_SIZE"] == 82
    assert insert_size_metrics_dict["PAIR_ORIENTATION"] == "FR"


@patch("builtins.open", return_value=StringIO(SAMTOOLS_STATS))
def test_parse_samtools_stats(mock_open):
    stats_dict = parsers.parse_samtools_stats("path")
    assert len(stats_dict) == 38
    assert "percentage of properly paired reads (%)" in stats_dict
    assert stats_dict["error rate"] == 0.003286209
    assert stats_dict["1st fragments"] == 137377422
